/*
 *  tty.c -- RS-232C o͊֘A
 *
 *  revision history:
 *	0.0: May  19, 1996 by Dai ISHIJIMA
 *	0.1: Jul. 27, 1996
 *	0.2: Aug. 16, 1996
 *	0.3: Aug. 24, 1996
 *	0.4: Dec.  2, 1998 (for Paldio Phone Book)
 *	0.5: Jan. 14, 1999 (for Chara-Mail Reader)
 */

#include <stdio.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/types.h>
#include <time.h>

#include <stdlib.h>
#include <unistd.h>

#ifndef O_NONBLOCK	/* for Sony NEWS, NEWS-OS 4.x */
#include <stream/xti.h>
#endif

#define FILEMODE 0777
#define TIMEOUT 15

#ifdef _POSIX_VDISABLE
#define CC_UNDEF _POSIX_VDISABLE
#else
#define CC_UNDEF 0
#endif

static struct termios t;
static struct termios new;


/* ǂݏpɃI[v */
int opentty(char *name)
{
    int fd;

    if ((fd = open(name, O_RDWR | O_NONBLOCK, FILEMODE)) <= 0) {
	fprintf(stderr, "can't open %s\n", name);
	exit(1);
    }
    return(fd);
}


/* VAx̃`FbN */
int speed(int sp)
{
    if (sp == 2400) {
	return(B2400);
    }
    else if (sp == 4800) {
	return(B4800);
    }
    else if (sp == 9600) {
	return(B9600);
    }
    else if (sp == 19200) {
	return(B19200);
    }
    else if (sp == 38400) {
	return(B38400);
    }
#ifdef B57600    
    else if (sp == 57600) {
	return(B57600);
    }
#endif
#ifdef B115200
    else if (sp == 115200) {
	return(B115200);
    }
#endif
    else {
	fprintf(stderr, "%d bps unavailable.  9600 bps selected.\n", sp);
    }
    return(B9600);
}


/* VA|[g̏Bx̐ݒERAW[h */
void ttyinit(int fd, long sp)
{
    int i;

    sp = speed(sp);
    tcgetattr(fd, &t);
    memcpy(&new, &t, sizeof(struct termios));
    new.c_iflag = IGNBRK | IGNPAR;
    new.c_oflag = 0;
    new.c_cflag = CS8 | CREAD | CLOCAL;
    new.c_lflag = 0;
    cfsetispeed(&new, sp);
    cfsetospeed(&new, sp);
    for (i = 0; i < NCCS; i++) {
	new.c_cc[i] = CC_UNDEF;
    }
#ifndef VMIN_NULL
    new.c_cc[VMIN] = t.c_cc[VMIN];
    new.c_cc[VTIME] = t.c_cc[VTIME];
#endif
    tcsetattr(fd, TCSANOW, &new);
}    


/* VA|[g̐ݒɖ߂ */
void ttyrestore(int fd)
{
    tcsetattr(fd, TCSANOW, &t);
}


/* VA|[g */
int closetty(int fd)
{
    ttyrestore(fd);
    return(close(fd));
}


/* 𑗐M */
/* 2000/07/14 T.Sagae oCif[^ unsigned t */
int ttyputc(int fd, unsigned char ch)
{
    int status;
    time_t t;

    t = time(NULL);
    while ((status = write(fd, &ch, 1)) < 1) {
#if 0	
	if (status < 0) {
	    fprintf(stderr, "ttyputc(): write failed\n");
	    return(EOF);
	}
#endif
	if (time(NULL) - t > TIMEOUT) {
	    fprintf(stderr, "ttyputc(): timeout\n");
	    return(EOF);
	}
    }
    return(ch);
}


/* MBMobt@̂Ƃ EOF 𑗂 */
int ttygetc(int fd)
{
    unsigned char buf[1];
    
    if (read(fd, &buf[0], 1) != 1) {
	return(EOF);
    }
    return(buf[0]);
}

#ifdef MAIN
/* MAIN`ĂƂ́Å֐`FbNvO𐶐 */

/*
 *  peeping-com -- Lady Godiva VA|[ĝ :-)
 *
 *  ̃vO COM1: (/dev/ttyd0) M
 *  COM2: (/dev/ttyd1) ɁACOM2: ̕ COM1: ɍđ
 *  ܂BRS-232C łƂ肳f[^̉͂ɍœK (?) łB
 *
 *  RpC@: gcc -DMAIN -o peeping-com tty.c
 *
 *  MAIN`ăRpCvOgł̐:
 *     EunbJ[ (hacker)vAƍߎ҂̈Ӗŗpl
 *       ̃vO𗘗p邱Ƃ֎~܂B
 *     Esgp (邢Ƃɂׂ炸 :-)
 *     Esgp (˂ɂׂ炸 :-)
 */

#define SPEED 9600		/* VȂx */
#define TTY0 "/dev/ttyd0"	/* COM1: ̃foCXt@C */
#define TTY1 "/dev/ttyd1"	/* COM2: ̃foCXt@C */
#define ONE2ZERO 1
#define ZERO2ONE 0
#define MAXQ 4096
#define NDUMP 16
#define DUMPTIMEOUT 5

#define shift --argc; ++argv
char *prog;
static int nq = 0;
static unsigned char queue[MAXQ];

#define dumpchar(ch) (((' ' <= (ch)) && ((ch) <= '~')) ? (ch) : '.')


void flushqueue(int dir)
{
    int i, j;

    for (i = 0; i < nq; i += NDUMP) {
	if (dir == ZERO2ONE) {
	    printf("0->1: ");
	}
	else if (dir == ONE2ZERO) {
	    printf("1->0: ");
	}
	for (j = 0; (j < NDUMP) && (i + j < nq); j++) {
	    printf("%02x ", queue[i + j]);
	    if (j == 7) {
		putchar(' ');
	    }
	}
	for (; j < NDUMP; j++) {
	    printf("   ");
	    if (j == 7) {
		putchar(' ');
	    }
	}
	putchar(' ');
	for (j = 0; (j < NDUMP) && (i + j < nq); j++) {
	    putchar(dumpchar(queue[i + j]));
	}
	putchar('\n');
	fflush(stdout);
    }
    nq = 0;
}


void enqueue(unsigned char ch, int dir)
{
    if (nq >= MAXQ) {
	flushqueue(dir);
    }
    queue[nq] = ch;
    ++nq;
}


void usage()
{
    fprintf(stderr, "Usage: %s -speed tty0 tty1\n", prog);
}


void main(int argc, char *argv[])
{
    int ch;
    int fd0, fd1;
    int dir;
    time_t t;
    char *tty0, *tty1;
    long speed;
    
    prog = *argv;
    shift;
    tty0 = TTY0;
    tty1 = TTY1;
    speed = SPEED;
    while ((argc > 0) && (argv[0][0] == '-')) {
	if (('0' <= argv[0][1]) && (argv[0][1] <= '9')) {
	    speed = -atoi(*argv);
	}
	else {
	    usage();
	    exit(1);
	}
	shift;
    }
    if (argc > 0) {
	tty0 = *argv;
	shift;
    }
    if (argc > 0) {
	tty1 = *argv;
	shift;
    }
    fd0 = opentty(tty0);
    ttyinit(fd0, speed);
    fd1 = opentty(tty1);
    ttyinit(fd1, speed);
    dir = -1;
    t = time(NULL);
    while (1) {
	if ((ch = ttygetc(fd0)) >= 0) {
	    t = time(NULL);
	    ttyputc(fd1, ch);
	    if (dir != ZERO2ONE) {
		flushqueue(dir);
	    }
	    dir = ZERO2ONE;
	    enqueue(ch, dir);
	}
	if ((ch = ttygetc(fd1)) >= 0) {
	    t = time(NULL);
	    ttyputc(fd0, ch);
	    if (dir != ONE2ZERO) {
		flushqueue(dir);
	    }
	    dir = ONE2ZERO;
	    enqueue(ch, dir);
	}
	if (time(NULL) - t > DUMPTIMEOUT) {
	    flushqueue(dir);
	    t = time(NULL);
	}
    }
    closetty(fd0); closetty(fd1);
}

/* Local Variables: */
/* compile-command:"cc -Wall -DMAIN -o peeping-com tty.c" */
/* End: */
#endif
